#!/usr/bin/python

import json
import sys
import getpass
import logging
import time
from os.path import expanduser

from efs.client.connection import Connection, RpcError
from efs.client.structure_file import StructureFile
from efs.client.files import Directory, SecurityError
from efs.client.mount import mount
from efs.common.encryption import import_key, export_key, generate_pubkey
from efs.client.keys import prompt_pubkey

class Client(object):

    def __init__(self, settings_file):
        self.settings_file = settings_file
        home = expanduser("~")
        self.session_file = home + "/.efs_session"
        with open(settings_file) as f:
            self.settings = json.load(f)

        self.conn = Connection(self.settings["host"],
                               self.settings["port"])

        if "server_public_key" in self.settings:
            if self.conn.server_public_key != import_key(self.settings["server_public_key"]):
                raise SecurityError("Server public key mismatch")
        else:
            prompt_pubkey("the server", self.conn.server_public_key)
            self.settings["server_public_key"] = export_key(self.conn.server_public_key)
            self.save_settings()

        self.structure = None

    def save_settings(self):
        with open(self.settings_file, "w") as f:
            json.dump(self.settings, f, indent=4)

    def save_session(self):
        session = {}
        session["host"] = self.settings["host"]
        session["port"] = self.settings["port"]
        session["token"] = self.conn.token
        session["user"] = self.conn.user
        session["session_key"] = self.conn.session_key
        session["private_key"] = export_key(self.conn.private_key)
        session["public_key"] = export_key(self.conn.public_key)
        with open(self.session_file, "w") as f:
            json.dump(session, f, indent=4)

    def register(self, user, pw):
        private_key = generate_pubkey()

        self.conn.private_key = private_key
        root_id = self.conn.register(user, pw, export_key(private_key.publickey()))
        self.conn.login(user, pw)
        
        self.structure = StructureFile(self.conn, fetch_data=False)
        home = Directory(self.structure, root_id, name=user, parent=self.structure.root)
        self.structure.root.children[user] = home
        self.structure.save()

        self.settings["keys"][user] = export_key(private_key, pw)
        self.save_settings()

        self.conn.mkgroup(user, user, private_key.publickey(), private_key)

    def login(self, user, pw):
        self.conn.private_key = import_key(self.settings["keys"][user], pw)
        self.conn.login(user, pw)

        self.structure = StructureFile(self.conn)

    def register_or_login(self, user, pw):
        try:
            self.login(user, pw)
        except:
            self.register(user, pw)

    def mount(self):
        mount(self.structure.root, self.settings["mountpoint"])

if __name__ == "__main__":
    if len(sys.argv) != 2:
        print >> sys.stderr, 'usage: %s SETTINGS_FILE' % sys.argv[0]
        exit(1)

    logging.getLogger().setLevel(logging.DEBUG)
    logging.getLogger().addHandler(logging.StreamHandler(sys.stderr))    

    client = Client(sys.argv[1])

    sys.stderr.write("Username: ")
    username = sys.stdin.readline()[:-1]
    password = getpass.getpass()
    client.register_or_login(username, password)
    client.save_session()
    client.mount()

    client.save_settings()
   
    
