import httplib, time

from efs.common.encryption import decrypt, encrypt, pubkey_decrypt, pubkey_encrypt, import_key, export_key
from efs.common.protocol import Command, Request, Response

class Connection(object):

    def __init__(self, host, port):
        self.conn = httplib.HTTPConnection(host, port)
        self.server_public_key = None
        self.private_key = None
        #Atttributes below are configured when the user logs in
        self.user = None
        self.token = None
        self.session_key = None
        self.public_key = None

        self.server_public_key = self.get_server_public_key()
    
    #Network methods
    def make_request(self, command, data):
        request = Request()
        request.user = self.user
        request.token = self.token
        request.command = command
        request.data = data
        request.timestamp = time.time()
        return request

    def simple_request(self, command_name, data, **kwargs):
        try:
            command = Command(command=command_name, args=kwargs)
            request = self.make_request(command, data)
            efs_response = self.send_request(request, True, self.session_key, "session:" + self.user)
            if efs_response.code == "Success":
                return efs_response.data
            else:
                raise RpcError(efs_response.code)
        except RpcError as e:
            raise RpcError(e.message)

    def send_request(self, request, encrypt_request, key=None, key_id=None):
        try:
            #Make request
            headers = {"Content-Type":"text/plain"}
            request_body = request.to_json()
            if encrypt_request == True:
                if key_id == "server":
                    request_body = pubkey_encrypt(request_body, self.server_public_key)
                else:
                    request_body = encrypt(request_body, self.session_key)
                headers["EFS-Encrypted"] = "true"
                headers["EFS-KeyId"] = key_id
            else:
                headers["EFS-Encrypted"] = "false"
            request_body = request_body.encode("utf-8")
            headers["Content-Length"] = len(request_body) 
            self.conn.request("POST", "", request_body, headers)

            #Process response
            response = self.conn.getresponse()
            if response.status == 200:
                efs_response_text = response.read().decode("utf-8")
                if response.getheader("EFS-Encrypted", "false") == "true":
                    user = self.user if self.user is not None else request.command.args["user"]
                    if response.getheader("EFS-KeyId", None) == "user:" + user:
                        efs_response_text = pubkey_decrypt(efs_response_text, self.private_key)
                    else:
                        efs_response_text = decrypt(efs_response_text, self.session_key)
                efs_response = Response.create_from_text(efs_response_text)
                return efs_response
            else: 
                raise RpcError("InvalidResponse")
        except:
            raise RpcError("InvalidResponse")

    #Key management methods
    def get_server_public_key(self):
        try:
            command = Command(command="get_server_public_key", args={})
            request = self.make_request(command, {})
            efs_response = self.send_request(request, False)
            if efs_response.code == "Success":
                self.server_public_key = import_key(efs_response.data['key'])
                return self.server_public_key
            else:
                raise RpcError(efs_response.code)
        except RpcError as e:
            raise RpcError(e.message)

    def get_user_public_key(self, user):
        try:
            data = self.simple_request("get_user_public_key", {}, user=user)
            return data["key"]
        except RpcError as e:
            raise RpcError(e.message)

    def get_group_public_key(self, group):
        try:
            data = self.simple_request("get_group_public_key", {}, group=group)
            return data["key"]
        except RpcError as e:
            raise RpcError(e.message)

    def get_group_private_key(self, group, user):
        try:
            data = self.simple_request("get_group_private_key", {}, group=group, user=user)
            return data["key"]
        except RpcError as e:
            raise RpcError(e.message)

    def get_decrypted_group_private_key(self, group):
        encrypted_key = self.get_group_private_key(group, self.user)
        return import_key(pubkey_decrypt(encrypted_key, self.private_key))

    #User and group management methods
    def login(self, user, password):
        try:
            args = {"user":user, "password":password}
            command = Command(command="login", args=args)
            request = self.make_request(command, {})
            efs_response = self.send_request(request, True, self.server_public_key, "server")
            if efs_response.code == "Success":
                self.user = user
                self.token = efs_response.data['token']
                self.session_key = efs_response.data['session_key']
                self.public_key = self.private_key.publickey()
            else:
                raise RpcError(efs_response.code)
        except RpcError as e:
            raise RpcError(e.message)

    def register(self, user, password, public_key):
        try:
            args = {"user":user, "password":password}
            data = {"public_key": public_key}
            command = Command(command="register", args=args)
            request = self.make_request(command, data)
            efs_response = self.send_request(request, True, self.server_public_key, "server")
            if efs_response.code == "Success":
                return efs_response.data["homedir"]
            else:
                raise RpcError(efs_response.code)
        except RpcError as e:
            raise RpcError(e.message)

    def chpasswd(self, user, new_password):
        try:
            self.simple_request("chpasswd", {}, user=user, new_password=new_password)
        except RpcError as e:
            raise RpcError(e.message)

    def mkgroup(self, group, owner, group_public_key, group_private_key):
        try:
            group_public_key = export_key(group_public_key)
            #Encrypt
            owner_private_key = pubkey_encrypt(export_key(group_private_key), self.public_key)
            self.simple_request("mkgroup", {},
                                group=group,
                                owner=owner,
                                group_public_key=group_public_key,
                                owner_private_key=owner_private_key)
        except RpcError as e:
            raise RpcError(e.message)
    
    def rmgroup(self, group):
        try:
            self.simple_request("rmgroup", {}, group=group)
        except RpcError as e:
            raise RpcError(e.message)

    def get_groups(self):
        try:
            data = self.simple_request("get_groups", {}, user=self.user)
            return data["groups"]
        except RpcError as e:
            raise RpcError(e.message)

    def adduser(self, group, user):
        try:
            user_public_key = import_key(self.get_user_public_key(user))
            group_private_key = self.get_group_private_key(group, self.user)
            group_private_key = pubkey_decrypt(group_private_key, self.private_key)
            member_private_key = pubkey_encrypt(group_private_key, user_public_key)
            self.simple_request("adduser", {"member_private_key":member_private_key}, group=group, user=user)
        except RpcError as e:
            raise RpcError(e.message)

    def rmuser(self, group, user):
        try:
            self.simple_request("rmuser", {}, group=group, user=user)
        except RpcError as e:
            raise RpcError(e.message)

    #File system methods

    def mknode(self, parent_id, perms):
        try:
            data = self.simple_request("mknode", {}, parent_id=parent_id, perms=perms)
            return data["node_id"]
        except RpcError as e:
            raise RpcError(e.message)
    
    def rmnode(self, node_id):
        try:
            self.simple_request("rmnode", {}, node_id=node_id)
        except RpcError as e:
            raise RpcError(e.message)
    
    def read(self, node_id):
        try:
            data = self.simple_request("read", {}, node_id=node_id)
            return data["contents"]
        except RpcError as e:
            raise RpcError(e.message)
    
    def write(self, node_id, contents):
        try:
            self.simple_request("write", {}, node_id=node_id, contents=contents)
        except RpcError as e:
            raise RpcError(e.message)

    def chmod(self, node_id, perms):
        try:
            self.simple_request("chmod", {}, node_id=node_id, perms=perms)
        except RpcError as e:
            raise RpcError(e.message)
    
    #only file owner can change the file group ownership
    def chgroup(self, node_id, group):
        try:
            # The group key is updated by chgrp in files.py, before it
            # calls this method

            self.simple_request("chgroup", node_id=node_id, group=group)
        except RpcError as e:
            raise RpcError(e.message)
    
    def mv(self, node_id, new_parent_id):
        try:
            self.simple_request("mv", {}, node_id=node_id, new_parent_id=new_parent_id)
        except RpcError as e:
            raise RpcError(e.message)

    def cp(self, node_id, copy_parent_id):
        try:
            data = self.simple_request("cp", {}, node_id=node_id, copy_parent_id=copy_parent_id)
            return data["new_node_id"]
        except RpcError as e:
            raise RpcError(e.message)

    def get_user_struct(self):
        try:
            data = self.simple_request("get_user_struct", {}, user=self.user)
            return data["struct"]
        except RpcError as e:
            raise RpcError(e.message)
    
    def update_user_struct(self, struct):
        try:
            self.simple_request("update_user_struct", {"struct":struct}, user=self.user)
        except RpcError as e:
            raise RpcError(e.message)

    def get_group_struct(self, group):
        try:
            data = self.simple_request("get_group_struct", {}, group=group)
            return data["struct"]
        except RpcError as e:
            raise RpcError(e.message)
    
    def update_group_struct(self, group, struct):
        try:
            self.simple_request("update_group_struct", {"struct":struct}, group=group)
        except RpcError as e:
            raise RpcError(e.message)

class RpcError(Exception):
    pass
